using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Drawing.Imaging;
using System.Text;
using System.Windows.Forms;
using System.IO;
using System.Runtime.InteropServices;
using System.Threading;

namespace Chip8 {
    public partial class MainInterface : Form {

        private MidiDevice SoundPlayer;

        public MainInterface(string[] Arguments) {

            UpdateColours(0x00, Color.White);
            UpdateColours(0xFF, Color.Black);

            Reset();
            KeyMap = new Dictionary<Keys, byte>();

            KeyMap.Add(Keys.D1, 0x1);
            KeyMap.Add(Keys.D2, 0x2);
            KeyMap.Add(Keys.D3, 0x3);
            KeyMap.Add(Keys.D4, 0xC);

            KeyMap.Add(Keys.Q, 0x4);
            KeyMap.Add(Keys.W, 0x5);
            KeyMap.Add(Keys.E, 0x6);
            KeyMap.Add(Keys.R, 0xD);

            KeyMap.Add(Keys.A, 0x7);
            KeyMap.Add(Keys.S, 0x8);
            KeyMap.Add(Keys.D, 0x9);
            KeyMap.Add(Keys.F, 0xE);

            KeyMap.Add(Keys.Z, 0xA);
            KeyMap.Add(Keys.X, 0x0);
            KeyMap.Add(Keys.C, 0xB);
            KeyMap.Add(Keys.V, 0xF);

            if (Arguments.Length != 0) {
                try {
                    LoadRom(Arguments[0].Trim('"'));
                    Reset();
                    Running = true;
                } catch { }
            }

            InitializeComponent();

            try {
                SoundPlayer = new MidiDevice();
                SoundEnabledOption.Enabled = true;
                SoundPlayer.SendMidiMessage(MidiDevice.MidiMessage.PatchChange, 0, 80);
            } catch {
                SoundPlayer = null;
                SoundEnabled = false;
            }

            int[] KeyNames = new int[] { 0x1, 0x2, 0x3, 0xC, 0x4, 0x5, 0x6, 0xD, 0x7, 0x8, 0x9, 0xE, 0xA, 0x0, 0xB, 0xF };

            for (int y = 0; y < 4; ++y) {
                for (int x = 0; x < 4; ++x) {
                    Button B = new Button();
                    
                    B.Dock = DockStyle.Fill;
                    B.MouseDown += new MouseEventHandler(ButtonClick);
                    B.MouseUp += new MouseEventHandler(ButtonRelease);
                    int i = x + y * 4;
                    B.Tag = KeyNames[i];
                    B.Text = KeyNames[i].ToString("X1");
                    ButtonGrid.Controls.Add(B, x, y);
                }
                
            }

        }

        void ButtonRelease(object sender, MouseEventArgs e) {
            KeyGrid[(int)((Button)sender).Tag] = false;
        }

        void ButtonClick(object sender, MouseEventArgs e) {
            KeyGrid[(int)((Button)sender).Tag] = true;
        }

        // CHIP-8 then SCHIP font data (copied to 0x000)
        private byte[] FontData = {
            0xF0, 0x90, 0x90, 0x90, 0xF0,
            0x20, 0x60, 0x20, 0x20, 0x70,
            0xF0, 0x10, 0xF0, 0x80, 0xF0,
            0xF0, 0x10, 0xF0, 0x10, 0xF0,
            0x90, 0x90, 0xF0, 0x10, 0x10,
            0xF0, 0x80, 0xF0, 0x10, 0xF0,
            0xF0, 0x80, 0xF0, 0x90, 0xF0,
            0xF0, 0x10, 0x20, 0x40, 0x40,
            0xF0, 0x90, 0xF0, 0x90, 0xF0,
            0xF0, 0x90, 0xF0, 0x10, 0xF0,
            0xF0, 0x90, 0xF0, 0x90, 0x90,
            0xE0, 0x90, 0xE0, 0x90, 0xE0,
            0xF0, 0x80, 0x80, 0x80, 0xF0,
            0xE0, 0x90, 0x90, 0x90, 0xE0,
            0xF0, 0x80, 0xF0, 0x80, 0xF0,
            0xF0, 0x80, 0xF0, 0x80, 0x80,

            0x7C, 0xC6, 0xCE, 0xDE, 0xD6, 0xF6, 0xE6, 0xC6, 0x7C, 0x00, 
            0x10, 0x30, 0xF0, 0x30, 0x30, 0x30, 0x30, 0x30, 0xFC, 0x00, 
            0x78, 0xCC, 0xCC, 0x0C, 0x18, 0x30, 0x60, 0xCC, 0xFC, 0x00, 
            0x78, 0xCC, 0x0C, 0x0C, 0x38, 0x0C, 0x0C, 0xCC, 0x78, 0x00, 
            0x0C, 0x1C, 0x3C, 0x6C, 0xCC, 0xFE, 0x0C, 0x0C, 0x1E, 0x00, 
            0xFC, 0xC0, 0xC0, 0xC0, 0xF8, 0x0C, 0x0C, 0xCC, 0x78, 0x00, 
            0x38, 0x60, 0xC0, 0xC0, 0xF8, 0xCC, 0xCC, 0xCC, 0x78, 0x00, 
            0xFE, 0xC6, 0xC6, 0x06, 0x0C, 0x18, 0x30, 0x30, 0x30, 0x00, 
            0x78, 0xCC, 0xCC, 0xEC, 0x78, 0xDC, 0xCC, 0xCC, 0x78, 0x00, 
            0x7C, 0xC6, 0xC6, 0xC6, 0x7C, 0x18, 0x18, 0x30, 0x70, 0x00, 
            0x30, 0x78, 0xCC, 0xCC, 0xCC, 0xFC, 0xCC, 0xCC, 0xCC, 0x00, 
            0xFC, 0x66, 0x66, 0x66, 0x7C, 0x66, 0x66, 0x66, 0xFC, 0x00, 
            0x3C, 0x66, 0xC6, 0xC0, 0xC0, 0xC0, 0xC6, 0x66, 0x3C, 0x00, 
            0xF8, 0x6C, 0x66, 0x66, 0x66, 0x66, 0x66, 0x6C, 0xF8, 0x00, 
            0xFE, 0x62, 0x60, 0x64, 0x7C, 0x64, 0x60, 0x62, 0xFE, 0x00, 
            0xFE, 0x66, 0x62, 0x64, 0x7C, 0x64, 0x60, 0x60, 0xF0, 0x00
        };

        // Maps a key (from the keypad) to a key number
        private Dictionary<Keys, byte> KeyMap;

        // Memory :)
        private byte[] MemoryMap;

        // ROM file once loaded
        private byte[] RomFile = new byte[0];

        private ushort CurrentInstruction;

        private byte CurrentImmediateValue {
            get { return (byte)(CurrentInstruction & 0xFF); }
        }

        // Registers (16 of 'em)
        private byte[] Registers;

        // Points to next instruction to be executed
        private ushort ProgramCounter;

        private ushort AddressRegister;

        // 16 keys: true for pressed, false for released
        private bool[] KeyGrid;

        // Switch for random mode
        private bool RandomIsAnd = true;

        // Was the sound playing last frame?
        private bool SoundOnLastFrame = false;

        // Call stack
        private Stack<ushort> CallStack;

        // Bitmapped display memory
        private bool[] DisplayMemory = new bool[128 * 64];

        // This is the 'real' 8-bit display storage
        private byte[] ExpandedDisplayMemory = new byte[128 * 64];

        // Used to blend pixels over a frame
        private int[] SmoothedDisplayMemory = new int[128 * 64];
        private uint FramesDrawn = 0;

        // Rate in Hz
        private int ClockSpeed = 800;

        // Don't mean anything in particular, but some games write to them so I've implemented them.
        // Data can be read/written to them, nothing else changes (NB: I support 16 for ease, only really 8 exist).
        private byte[] Hp48Flags;

        // Source of random numbers (what else?)
        private Random RandomNumberSource;

        // Switches for wrapping sprites in X or Y axes
        private bool SpriteXWrap = true;
        private bool SpriteYWrap = true;

        // Sound on/off?
        private bool SoundEnabled = true;

        // The following properties make life easier ("virtual" registers)

        private byte RegisterX {
            get { return Registers[(CurrentInstruction >> 8) & 0xF]; }
            set { Registers[(CurrentInstruction >> 8) & 0xF] = value; }
        }

        private byte RegisterY {
            get { return Registers[(CurrentInstruction >> 4) & 0xF]; }
            set { Registers[(CurrentInstruction >> 4) & 0xF] = value; }
        }

        private byte RegisterFlag {
            get { return Registers[0xF]; }
            set { Registers[0xF] = value; }
        }

        // Are we running?
        private bool Running = false;

        // The two harware timers
        private byte DelayTimer;
        private byte SoundTimer;


        // SCHIP or CHIP-8 mode?
        
        private enum DisplayMode {
            Chip8, SChip
        }

        private DisplayMode CurrentDisplayMode;

        // Screen bitmap

        private Bitmap DisplayBitmap = new Bitmap(128, 64, PixelFormat.Format8bppIndexed);

        // Time of the last frame (set to -1 to force an update).
        DateTime LastFrame = DateTime.Now.Subtract(new TimeSpan(0, 0, 1));

        // Clear everything, copy ROM/font in, reset timers &c
        private void Reset() {
            Registers = new byte[16];
            AddressRegister = 0x0000;
            ProgramCounter = 0x0200;
            MemoryMap = new byte[Math.Max(0x1000, RomFile.Length + 0x0200)];
            RomFile.CopyTo(MemoryMap, 0x0200);
            FontData.CopyTo(MemoryMap, 0x0000);
            CallStack = new Stack<ushort>();
            CurrentDisplayMode = DisplayMode.Chip8;
            RandomNumberSource = new Random();
            KeyGrid = new bool[16];
            Hp48Flags = new byte[16];
            ClearScreen();
            
            SoundOnLastFrame = false;
            DelayTimer = 0;
            SoundTimer = 0;
            LastFrame = DateTime.Now;

            if (SoundPlayer != null) SoundPlayer.SendMidiMessage(MidiDevice.MidiMessage.KeyRelease, 0, 80);
        }

        private void LoadRom(string filename) {
            using (BinaryReader BR = new BinaryReader(new FileStream(filename, FileMode.Open))) {
                this.RomFile = BR.ReadBytes((int)BR.BaseStream.Length);
            }
        }

        private void ClearScreen() {
            for (int i = 0; i < DisplayMemory.Length; ++i) DisplayMemory[i] = false;
            ScreenImageChanged();
        }

        private void ScreenImageChanged() {
            for (int i = 0; i < 128 * 64; ++i) {
                SmoothedDisplayMemory[i] += DisplayMemory[i] ? 1 : 0;
            }
            ++FramesDrawn;
        }

        // Fetch and execute a single instruction
        private void FetchExecute() {
            if (ProgramCounter < 0 || ProgramCounter >= MemoryMap.Length - 1) throw new Exception("Program counter has gone out of bounds.");
            CurrentInstruction = (ushort)((MemoryMap[ProgramCounter++] << 8) | MemoryMap[ProgramCounter++]);

            switch (CurrentInstruction & 0xF000) {
                case 0x0000:
                    switch (CurrentInstruction & 0xFF) {
                        case 0xE0:
                            ClearScreen();
                            break;
                        case 0xFE:
                            CurrentDisplayMode = DisplayMode.Chip8;
                            break;
                        case 0xFF:
                            CurrentDisplayMode = DisplayMode.SChip;
                            break;
                        case 0xEE:
                            if (CallStack.Count < 1) throw new Exception("Stack underflow.");
                            ProgramCounter = CallStack.Pop();
                            break;
                        case 0xFC: {
                                int p = 0;
                                for (int y = 0; y < 64; ++y) {
                                    for (int x = 0; x < 128; ++x) {
                                        if (x < 124) {
                                            DisplayMemory[p] = DisplayMemory[p + 4];
                                        } else {
                                            DisplayMemory[p] = false;
                                        }
                                        ++p;
                                    }
                                }
                            }
                            ScreenImageChanged();
                            break;
                        case 0xFB: {
                                int p = 128 * 64 - 1;
                                for (int y = 0; y < 64; ++y) {                                    
                                    for (int x = 127; x >= 0; --x) {
                                        if (x >= 4) {
                                            DisplayMemory[p] = DisplayMemory[p - 4];
                                        } else {
                                            DisplayMemory[p] = false;
                                        }
                                        --p;
                                    }
                                }
                            }
                            ScreenImageChanged();
                            break;
                        case 0xFD:
                            ProgramCounter -= 2;
                            break;
                        default:
                            if ((CurrentInstruction & 0x00F0) != 0x00C0) throw new Exception("Unsupported instruction " + CurrentInstruction.ToString("X4") + ".");
                            int LinesToScroll = CurrentInstruction & 0xF;
                            for (int i = 0; i < LinesToScroll; ++i) {
                                for (int j = 8191; j >= 0; --j) {
                                    if (j < 128) {
                                        DisplayMemory[j] = false;
                                    } else {
                                        DisplayMemory[j] = DisplayMemory[j - 128];
                                    }
                                }   
                            }
                            ScreenImageChanged();
                            break;
                    }
                    break;
                case 0x1000:
                case 0x2000:
                    if ((CurrentInstruction & 0xF000) == 0x2000)CallStack.Push(ProgramCounter);
                    ProgramCounter = (ushort)(CurrentInstruction & 0x0FFF);
                    break;
                case 0x3000:
                    if (RegisterX == CurrentImmediateValue) ProgramCounter += 2;
                    break;
                case 0x4000:
                    if (RegisterX != CurrentImmediateValue) ProgramCounter += 2;
                    break;
                case 0x5000:
                    if (RegisterX == RegisterY) ProgramCounter += 2;
                    break;
                case 0x6000:
                    RegisterX = CurrentImmediateValue;
                    break;
                case 0x7000:
                    RegisterX += CurrentImmediateValue;
                    break;
                case 0x8000:
                    switch (CurrentInstruction & 0x000F) {
                        case 0x0:
                            RegisterX = RegisterY;
                            break;
                        case 0x1:
                            RegisterX |= RegisterY;
                            break;
                        case 0x2:
                            RegisterX &= RegisterY;
                            break;
                        case 0x3:
                            RegisterX ^= RegisterY;
                            break;
                        case 0x4:
                            RegisterFlag = (byte)((RegisterX + RegisterY) > 0xFF ? 1 : 0);
                            RegisterX += RegisterY;
                            break;
                        case 0x5:
                            RegisterFlag = (byte)(RegisterX >= RegisterY ? 1 : 0);
                            RegisterX -= RegisterY;
                            break;
                        case 0x6:
                            RegisterFlag = (byte)(RegisterX & 0x01);
                            RegisterX >>= 1;
                            break;
                        case 0x7:
                            RegisterFlag = (byte)(RegisterY >= RegisterX ? 1 : 0);
                            RegisterX = (byte)(RegisterY - RegisterX);
                            break;
                        case 0xE:
                            RegisterFlag = (byte)((RegisterX & 0x80) >> 7);
                            RegisterX <<= 1;
                            break;
                        default:
                            throw new Exception("Unsupported instruction " + CurrentInstruction.ToString("X4") + ".");
                    }
                    break;
                case 0x9000:
                    if (RegisterX != RegisterY) ProgramCounter += 2;
                    break;
                case 0xA000:
                    AddressRegister = (ushort)(CurrentInstruction & 0x0FFF);
                    break;
                case 0xB000:
                    ProgramCounter = (ushort)(CurrentInstruction & 0x0FFF + Registers[0]);
                    break;
                case 0xC000:
                    RegisterX = (byte)(RandomIsAnd ? RandomNumberSource.Next() & CurrentImmediateValue : RandomNumberSource.Next() % CurrentImmediateValue);
                    break;
                case 0xD000:
                    
                    RegisterFlag = 0;
                    int SpriteWidth = 8;
                    int SpriteHeight = CurrentInstruction & 0xF;

                    int SpriteX = RegisterX;
                    int SpriteY = RegisterY;

                    int MaxX = 127;
                    int MaxY = 63;

                    if (CurrentDisplayMode == DisplayMode.Chip8) {
                        MaxX = 63;
                        MaxY = 31;
                    }

                    if (SpriteHeight == 0) {
                        switch (CurrentDisplayMode) {
                            case DisplayMode.Chip8:
                                SpriteHeight = 16;
                                break;
                            case DisplayMode.SChip:
                                SpriteHeight = 16;
                                SpriteWidth = 16;
                                break;
                        } 
                        
                    }
                    int SpriteRowPointer = AddressRegister;
                    for (int y = 0; y < SpriteHeight; ++y) {

                        if (y + SpriteY > MaxY && !SpriteYWrap) break;

                        byte PixelRow = MemoryMap[SpriteRowPointer];
                        for (int x = 0; x < SpriteWidth; ++x) {

                            if (x + SpriteX > MaxX && !SpriteXWrap) break;


                            if (x == 8) {
                                PixelRow = MemoryMap[++SpriteRowPointer];
                            }

                            bool SetPixel = (PixelRow & 0x80) != 0x00;
                            PixelRow <<= 1;
                            if (!SetPixel) continue;

                            int PixelX = x + SpriteX;
                            int PixelY = y + SpriteY;
                            if (CurrentDisplayMode == DisplayMode.Chip8) {
                                PixelX *= 2;
                                PixelY *= 2;
                            }
                            int PixelPointer = (PixelX & 127) + (PixelY & 63) * 128;
                            if (DisplayMemory[PixelPointer]) {
                                RegisterFlag = 1;
                            }
                                DisplayMemory[PixelPointer] ^= true;
                            if (CurrentDisplayMode == DisplayMode.Chip8) {
                                DisplayMemory[PixelPointer + 0x01] = DisplayMemory[PixelPointer];
                                DisplayMemory[PixelPointer + 0x80] = DisplayMemory[PixelPointer];
                                DisplayMemory[PixelPointer + 0x81] = DisplayMemory[PixelPointer];
                            }
                        }
                        ++SpriteRowPointer;
                    }
                    ScreenImageChanged();
                    break;
                case 0xE000:
                    switch (CurrentInstruction & 0xFF) {
                        case 0x9E:
                            if (KeyGrid[RegisterX]) ProgramCounter += 2;
                            break;
                        case 0xA1:
                            if (!KeyGrid[RegisterX]) ProgramCounter += 2;
                            break;
                        default:
                            throw new Exception("Unsupported instruction " + CurrentInstruction.ToString("X4") + ".");
                    }
                    break;
                case 0xF000:
                    switch (CurrentInstruction & 0xFF) {
                        case 0x07:
                            RegisterX = DelayTimer;
                            break;
                        case 0x0A:
                            bool KeyFound = false;
                            for (byte i = 0; i < 16; ++i) {
                                if (KeyGrid[i]) {
                                    RegisterX = i;
                                    KeyFound = true;
                                    break;
                                }
                            }
                            if (!KeyFound) ProgramCounter -= 2;
                            break;
                        case 0x15:
                            DelayTimer = RegisterX;
                            break;
                        case 0x18:
                            SoundTimer = RegisterX;
                            break;
                        case 0x1E:
                            AddressRegister += RegisterX;
                            break;
                        case 0x29:
                            AddressRegister = (byte)(RegisterX * 5);
                            break;
                        case 0x30:
                            AddressRegister = (byte)(RegisterX * 10 + 5 * 16);
                            break;
                        case 0x33:
                            int ToExpand = RegisterX;
                            MemoryMap[AddressRegister + 0] = (byte)(ToExpand / 100); ToExpand -= MemoryMap[AddressRegister + 0] * 100;
                            MemoryMap[AddressRegister + 1] = (byte)(ToExpand / 10); ToExpand -= MemoryMap[AddressRegister + 1] * 10;
                            MemoryMap[AddressRegister + 2] = (byte)(ToExpand);
                            break;
                        case 0x55:
                            for (int i = 0; i <= (CurrentInstruction & 0x0F00) >> 8; ++i) MemoryMap[AddressRegister + i] = Registers[i];                                
                            break;
                        case 0x65:
                            for (int i = 0; i <= (CurrentInstruction & 0x0F00) >> 8; ++i) Registers[i] = MemoryMap[AddressRegister + i];
                            break;
                        case 0x75:
                            for (int i = 0; i <= (CurrentInstruction & 0x0F00) >> 8; ++i) Hp48Flags[i] = Registers[i];
                            break;
                        case 0x85:
                            for (int i = 0; i <= (CurrentInstruction & 0x0F00) >> 8; ++i) Registers[i] = Hp48Flags[AddressRegister];
                            break;
                        default:
                            throw new Exception("Unsupported instruction " + CurrentInstruction.ToString("X4") + ".");
                    }
                    break;
                default:
                    throw new Exception("Unsupported instruction " + CurrentInstruction.ToString("X4") + ".");
            }
        }

        // Write to the display bitmap
        private void RefreshDisplay() {
            if (FramesDrawn == 0) FramesDrawn++;
            
            for (int i = 0; i < 128 * 64; ++i) {
                ExpandedDisplayMemory[i] = (byte)Math.Min(0xFF, (SmoothedDisplayMemory[i] * 256) / FramesDrawn);
                SmoothedDisplayMemory[i] = 0;
            }
            ScreenImageChanged();
            FramesDrawn = 0;

            BitmapData BD = DisplayBitmap.LockBits(new Rectangle(0, 0, 128, 64), ImageLockMode.WriteOnly, PixelFormat.Format8bppIndexed);
            Marshal.Copy(ExpandedDisplayMemory, 0, BD.Scan0, 128 * 64);
            DisplayBitmap.UnlockBits(BD);

        }

        
        

        private void MainInterface_Load(object sender, EventArgs e) {

            this.Show();
            this.Focus();
            

            DateTime LastSpeedCheck = DateTime.Now.Subtract(new TimeSpan(0, 0, 1));
            DateTime LastDelay = DateTime.Now.Subtract(new TimeSpan(0, 0, 1));
            uint InstructionCount = 0;

            Graphics RenderOut = DisplayWindow.CreateGraphics();

            while (this.Visible) {
                
                RefreshDisplay();

                DisplayWindow.Refresh();
                Application.DoEvents();

                DateTime ThisFrame = DateTime.Now;
                TimeSpan TimeForFrame = ThisFrame - LastFrame;


                
                if (Running) {                    

                    try {
                        for (int i = 0; i < (int)Math.Round(TimeForFrame.TotalSeconds * ClockSpeed, 0); ++i) {
                            FetchExecute();
                            ++InstructionCount;
                        }
                    } catch (Exception ex) {
                        MessageBox.Show(this, ex.Message, "Emulation Error", MessageBoxButtons.OK, MessageBoxIcon.Information);
                        Running = false;
                    }
                    ThisFrame = DateTime.Now;

                    if (SoundEnabled && SoundPlayer != null) {
                        if (SoundTimer != 0 && !SoundOnLastFrame) {
                            SoundPlayer.SendMidiMessage(MidiDevice.MidiMessage.KeyPress, 0, 80, 127);
                            SoundOnLastFrame = true;
                        } else if (SoundTimer == 0 && SoundOnLastFrame) {
                            SoundPlayer.SendMidiMessage(MidiDevice.MidiMessage.KeyRelease, 0, 80);
                            SoundOnLastFrame = false;
                        }
                    }

                    while (ThisFrame >= LastDelay + new TimeSpan(166667)) {
                        if (DelayTimer > 0) --DelayTimer;
                        if (SoundTimer > 0) --SoundTimer;
                        LastDelay += new TimeSpan(166667);
                    }
                    
                }

                if (ThisFrame >= LastSpeedCheck + new TimeSpan(0, 0, 1)) {
                    double Ips = (double)InstructionCount / (ThisFrame.Subtract(LastSpeedCheck)).TotalSeconds;
                    double Percentage = Ips / (double)this.ClockSpeed;
                    this.InstructionPerSecondCount.Text = Ips.ToString("N") + "Hz (" + Percentage.ToString("P0") + ")";
                    InstructionCount = 0;
                    LastSpeedCheck = ThisFrame;
                }

                LastFrame = ThisFrame;
            }
        }

        private void ChangeKeyState(Keys keyChanged, bool isDown) {
            byte KeyCode;
            if (KeyMap.TryGetValue(keyChanged, out KeyCode)) {
                KeyGrid[KeyCode] = isDown;
            }
        }

        private void MainInterface_KeyUp(object sender, KeyEventArgs e) {
            ChangeKeyState(e.KeyCode, false);
        }

        private void MainInterface_KeyDown(object sender, KeyEventArgs e) {
            ChangeKeyState(e.KeyCode, true);
        }

        private void resetToolStripMenuItem_Click(object sender, EventArgs e) {
            Reset();
        }

        private void loadROMToolStripMenuItem_Click(object sender, EventArgs e) {
            if (OpenRomDialog.ShowDialog() == DialogResult.OK) {
                try {
                    LoadRom(OpenRomDialog.FileName);
                    Reset();
                    Running = true;
                } catch (Exception ex) {
                    MessageBox.Show(this, "Error loading file: \n" + ex.Message, "Load ROM", MessageBoxButtons.OK, MessageBoxIcon.Error);
                }
            }
        }

        private void debugToolStripMenuItem_DropDownOpening(object sender, EventArgs e) {
            this.runToolStripMenuItem.Checked = Running;
        }

        private void runToolStripMenuItem_Click(object sender, EventArgs e) {
            Running ^= true;
        }

        
        private void DrawColourPreviewBox(ToolStripMenuItem preview, Color colour) {
            Graphics g = Graphics.FromImage(preview.Image);
            g.Clear(colour);
            g.DrawRectangle(Pens.Black, 0, 0, 15, 15);
            preview.Invalidate();
        }

        private void setDisplayColoursToolStripMenuItem_DropDownOpening(object sender, EventArgs e) {
            if (PixelOffColour.Image == null) PixelOffColour.Image = new Bitmap(16, 16);
            if (PixelOnColour.Image == null) PixelOnColour.Image = new Bitmap(16, 16);            

            DrawColourPreviewBox(PixelOffColour, this.DisplayBitmap.Palette.Entries[0x00]);
            DrawColourPreviewBox(PixelOnColour, this.DisplayBitmap.Palette.Entries[0xFF]);

        }
        private void UpdateColours(int index) {
            UpdateColours(index, null);
        }

        // Fills the palette from 0x00->0xFF, interpolating as it goes

        private void UpdateColours(int index, Color? colour) {
            if (!colour.HasValue) {
                SetColourDialog.Color = this.DisplayBitmap.Palette.Entries[index];
                if (SetColourDialog.ShowDialog() == DialogResult.Cancel) return;
                colour = SetColourDialog.Color;
            }
            ColorPalette p = this.DisplayBitmap.Palette;
            p.Entries[index] = Color.FromArgb(0xFF, colour.Value);
            Color A = p.Entries[0x00];
            Color B = p.Entries[0xFF];

            for (int i = 1; i <= 254; ++i) {
                int CR = (int)((double)A.R + (double)(B.R - A.R) * (double)i / 256);
                int CG = (int)((double)A.G + (double)(B.G - A.G) * (double)i / 256);
                int CB = (int)((double)A.B + (double)(B.B - A.B) * (double)i / 256);
                p.Entries[i] = Color.FromArgb(0xFF, CR, CG, CB);
            }

            this.DisplayBitmap.Palette = p;
        }

        private void PixelOffColour_Click(object sender, EventArgs e) {
            UpdateColours(0x00);
        }

        private void PixelOnColour_Click(object sender, EventArgs e) {
            UpdateColours(0xFF);
        }

        private void swapToolStripMenuItem_Click(object sender, EventArgs e) {
            Color A = this.DisplayBitmap.Palette.Entries[0x00];
            Color B = this.DisplayBitmap.Palette.Entries[0xFF];
            UpdateColours(0x00, B);
            UpdateColours(0xFF, A);
        }

        private void SpeedPreview_TextChanged(object sender, EventArgs e) {
  
        }

        private void SetSpeed_DropDownOpening(object sender, EventArgs e) {
            SpeedPreview.Text = ClockSpeed.ToString();
        }

        private void exitToolStripMenuItem_Click(object sender, EventArgs e) {
            this.Close();
        }

        private void SpriteXWrapOption_Click(object sender, EventArgs e) {
            SpriteXWrap ^= true;
        }

        private void SpriteYWrapOption_Click(object sender, EventArgs e) {
            SpriteYWrap ^= true;
        }

        private void emulationToolStripMenuItem_DropDownOpening(object sender, EventArgs e) {
            SpriteXWrapOption.Checked = SpriteXWrap;
            SpriteYWrapOption.Checked = SpriteYWrap;
            
        }

        private void SoundEnabledOption_Click(object sender, EventArgs e) {
            SoundEnabled ^= true;
            if (!SoundEnabled) SoundPlayer.SendMidiMessage(MidiDevice.MidiMessage.KeyRelease, 0, 80);
        }

        private void RandomModeOption_DropDownOpening(object sender, EventArgs e) {
            RandomAnd.Checked = RandomIsAnd;
            RandomMod.Checked = !RandomIsAnd;
        }

        private void RandomAnd_Click(object sender, EventArgs e) {
            RandomIsAnd = true;
        }

        private void RandomMod_Click(object sender, EventArgs e) {
            RandomIsAnd = false;
        }

        private void optionsToolStripMenuItem_DropDownOpening(object sender, EventArgs e) {
            SoundEnabledOption.Checked = SoundEnabled;
            smoothScalingToolStripMenuItem.Checked = SmoothScaling;
            showKeypadToolStripMenuItem.Checked = KeyPadPanel.Visible;
        }


        private void SpeedPreview_KeyUp(object sender, KeyEventArgs e) {
            if (e.KeyCode != Keys.Enter) return;
            int OldSpeed = ClockSpeed;
            if (!int.TryParse(SpeedPreview.Text, out ClockSpeed) || ClockSpeed < 1) {
                SpeedPreview.Text = OldSpeed.ToString();
            }
        }

        private bool SmoothScaling = false;
        private void DisplayWindow_Paint(object sender, PaintEventArgs e) {
            e.Graphics.InterpolationMode = SmoothScaling ? System.Drawing.Drawing2D.InterpolationMode.Bilinear : System.Drawing.Drawing2D.InterpolationMode.NearestNeighbor;
            e.Graphics.PixelOffsetMode = System.Drawing.Drawing2D.PixelOffsetMode.Half;
            e.Graphics.DrawImage(DisplayBitmap, DisplayWindow.ClientRectangle);
        }

        private void smoothScalingToolStripMenuItem_Click(object sender, EventArgs e) {
            SmoothScaling ^= true;
        }


        private void showKeypadToolStripMenuItem_Click(object sender, EventArgs e) {
            this.KeyPadPanel.Visible ^= true;
        }

        private void hzToolStripMenuItem_Click(object sender, EventArgs e) {
            ClockSpeed += 200;
        }

        private void hzToolStripMenuItem1_Click(object sender, EventArgs e) {
            ClockSpeed = Math.Max(100, ClockSpeed - 200);
        }

    }
}